/*
 *  PHIGS.c
 *  PHIGS
 *
 *  Created by Simone Portuesi on 15/03/06.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */

/* OpenGL implementation of PHIGS Orientation and View Mapping  */

#include "PHIGS.h"
#include "utility.h"

/* Include GL, GLU & GLUT */
#ifdef WIN32
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h>
#endif

#ifdef linux
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

int PHIGS_use_Gl=0;

/* Implements PHIGS View Orientation in PHIGS style */
void PHIGS_ViewOrientation(struct PHIGS_ViewOrientation *orientation) {

	/* Model Matrix */
	glMatrixMode(GL_MODELVIEW);
	
	/* Load identity */
	glLoadIdentity();
	
	if(PHIGS_use_Gl) {/* Try to use GL */
		double center[3];
	
		/* center = vrp - vpn */
		VECT_OP_VECT(center,-,orientation->vrp,orientation->vpn);
	
		/* Look at */
		gluLookAt(orientation->vrp[0],orientation->vrp[1],orientation->vrp[2],
				  center[0],center[1],center[2],
				  orientation->vuv[0],orientation->vuv[1],orientation->vuv[2]);
				  
	} else { /* Do everythin by hand */
		double rn[3],ru[3],rv[3];
		double matrix[4][4]=      { { 1, 0, 0, 0 },
									{ 0, 1, 0, 0 },
									{ 0, 0, 1, 0 },
									{ 0, 0, 0, 1 } };
		double norm;
		
		/* Rn = VPN / |VPN| */
		norm=NORM_VECT(orientation->vpn);
	
		VECT_OP_SCAL(rn,/,orientation->vpn,norm);
	
		/* Ru = VUV x RN */
		/* Ru = Ru / |Ru | */
		CROSSPROD(ru,orientation->vuv,rn);
	
		norm=NORM_VECT(ru);
	
		VECT_OP_SCAL(ru,/,ru,norm);
	
		/* Rv = Rn x Ru */
		CROSSPROD(rv,rn,ru);
	
		/* Insert Ru, Rv, Rn in Rotation Matrix */
		matrix[0][0]=ru[0];
		matrix[1][0]=ru[1];
		matrix[2][0]=ru[2];
		matrix[0][1]=rv[0];
		matrix[1][1]=rv[1];
		matrix[2][1]=rv[2];
		matrix[0][2]=rn[0];
		matrix[1][2]=rn[1];
		matrix[2][2]=rn[2];
	
		/* Rotate ° Translate(-VRP) */
		glMultMatrixd(matrix);
		glTranslated(-orientation->vrp[0],-orientation->vrp[1],-orientation->vrp[2]);
	}			  
}


/* Implements PHIGS View Map in PHIGS */
void PHIGS_ViewMapping(struct PHIGS_ViewMapping *view) {
	double DOP[3];
	double SHpar[4][4]=   { { 1, 0, 0, 0 },
							{ 0, 1, 0, 0 },
							{ 0, 0, 1, 0 },
							{ 0, 0, 0, 1 } };
						   
	/* DOP = Center BOX - PRP */
	DOP[0] = ( view->box.u_max + view->box.u_min ) / 2 - view->prp[0];
	DOP[1] = ( view->box.v_max + view->box.v_min ) / 2 - view->prp[1];
	DOP[2] = -view->prp[2];
			 
	/* Shear factors SHx = DOPx / DOPz, Shy= DOPy / DOPz */
	SHpar[2][0]= - DOP[0] / DOP[2];
	SHpar[2][1]= - DOP[1] / DOP[2];

	/* Projection Matrix */
	glMatrixMode(GL_PROJECTION);
	
	glLoadIdentity();
	
	if(view->prjType==PARALLEL) {
		if(PHIGS_use_Gl) { /* Use GL prim */
			/* Just do it (need to compose SHEAR) */
			glOrtho( view->box.u_min, view->box.u_max, view->box.v_min, view->box.v_max, -view->box.front, -view->box.back );
		} else {
			/* Set default viev voume, (GL clip planes) */
			glOrtho( -1, 1, -1, 1, -1, 1);
			
			/* Spar ~ Tpar */
			glTranslated( - ( view->box.u_min + view->box.u_max ) / 2,
						  - ( view->box.v_min + view->box.v_max ) / 2,
						  - ( view->box.front + view->box.back  ) / 2
						  );	
		
			glScaled( 2 / ( view->box.u_max - view->box.u_min ),
					  2 / ( view->box.v_max - view->box.v_min ),
				      2 / ( view->box.front - view->box.back  ) 
					  					  );
		}
		
		/* Multiply (post compose) shear matrix */
		glMultMatrixd(SHpar);		  

	} else { /* TBD */
		double vrp_prime[3];
	
		/* VRP' = SHpar ~ T(-PRP) ( [0, 0, 0, 1] ) */
		vrp_prime[0]=-view->prp[0] - SHpar[2][0] * view->prp[2]; 
		vrp_prime[1]=-view->prp[1] - SHpar[2][1] * view->prp[2];
		vrp_prime[2]=-view->prp[2];

		/* Set up "default" frustum ( cono a base poligonale troncato) */
		glFrustum(-1,1,-1,1,1,3);
		
		/* Place front plane on z= -1 */
		glScaled( - 2 / ( vrp_prime[2] - view->box.front ),
		          - 2 / ( vrp_prime[2] - view->box.front ),
		          - 2 / ( vrp_prime[2] - view->box.front )
				);
        
		/* Make unit with unit scope */
		glScaled( ( - 2 * vrp_prime[2] ) / ( view->box.u_max - view->box.u_min ),
				  ( - 2 * vrp_prime[2] ) / ( view->box.v_max - view->box.v_min ),
				  1);

		/* Shear matrix "straighting" the frustrum */
		glMultMatrixd(SHpar);	

	    /* Translate in center of projection */
		glTranslated( -view->prp[0],
					  -view->prp[1],
					  -view->prp[2]);		}
	
	/* Back to model matrix */
	glMatrixMode(GL_MODELVIEW);

}

/* Implements both orientation and map view in PHIGS style */
void PHIGS_View(struct PHIGS_View *view) {
	
	PHIGS_ViewMapping(&(view->mapping));
	PHIGS_ViewOrientation(&(view->orientation));
	
}